@@ -14,11 +14,17 @@
 # limitations under the License.
 #
 
+from typing import TYPE_CHECKING, Optional
+
 from pyspark import SparkContext
 from pyspark.sql import utils
 from pyspark.sql.utils import CapturedException
 
 
+if TYPE_CHECKING:
+    from py4j.java_gateway import JavaObject, JVMView  # type: ignore[import]
+
+
 class DeltaConcurrentModificationException(CapturedException):
     """
     The basic class for all Delta commit conflict exceptions.
@@ -105,13 +111,15 @@ class ConcurrentTransactionException(CapturedException):
 _delta_exception_patched = False
 
 
-def _convert_delta_exception(e):
+def _convert_delta_exception(e: "JavaObject") -> Optional[CapturedException]:
     """
     Convert Delta's Scala concurrent exceptions to the corresponding Python exceptions.
     """
-    s = e.toString()
-    c = e.getCause()
-    stacktrace = SparkContext._jvm.org.apache.spark.util.Utils.exceptionString(e)
+    s: str = e.toString()
+    c: "JavaObject" = e.getCause()
+
+    jvm: "JVMView" = SparkContext._jvm  # type: ignore[attr-defined]
+    stacktrace = jvm.org.apache.spark.util.Utils.exceptionString(e)
 
     if s.startswith('io.delta.exceptions.DeltaConcurrentModificationException: '):
         return DeltaConcurrentModificationException(s.split(': ', 1)[1], stacktrace, c)
@@ -132,14 +140,14 @@ def _convert_delta_exception(e):
     return None
 
 
-def _patch_convert_exception():
+def _patch_convert_exception() -> None:
     """
     Patch PySpark's exception convert method to convert Delta's Scala concurrent exceptions to the
     corresponding Python exceptions.
     """
     convert_sql_exception = utils.convert_exception
 
-    def convert_delta_exception(e):
+    def convert_delta_exception(e: "JavaObject") -> Optional[CapturedException]:
         delta_exception = _convert_delta_exception(e)
         if delta_exception is not None:
             return delta_exception
